# Generated by Django 2.1.7 on 2019-04-23 12:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("nuntius", "0012_campaignsentevent_tracking_id")]

    operations = [
        migrations.AlterField(
            model_name="campaignsentevent",
            name="esp_message_id",
            field=models.CharField(
                editable=False,
                max_length=255,
                null=True,
                unique=True,
                verbose_name="ID given by the sending server",
            ),
        )
    ]



import logging

from django.db import transaction
from django.dispatch import receiver

from nuntius.actions import update_subscriber
from nuntius.admin import subscriber_class
from nuntius.models import CampaignSentEvent, CampaignSentStatusType

logger = logging.getLogger(__name__)

try:
    from anymail.signals import tracking, EventType
except ImportError:
    pass
else:
    actions = {
        EventType.SENT: CampaignSentStatusType.OK,
        EventType.REJECTED: CampaignSentStatusType.REJECTED,
        EventType.FAILED: CampaignSentStatusType.ERROR,
        EventType.BOUNCED: CampaignSentStatusType.BOUNCED,
        EventType.UNSUBSCRIBED: CampaignSentStatusType.UNSUBSCRIBED,
        EventType.COMPLAINED: CampaignSentStatusType.COMPLAINED,
    }

    @receiver(tracking, dispatch_uid="nuntius_anymail_tracking")
    def handle_anymail(sender, event, esp_name, **kwargs):
        campaign_status = actions.get(event.event_type, (None, None))

        if event.event_type == EventType.BOUNCED:
            # in cases of soft bounces (grouped by Anymail with hard bounces in the BOUNCED event type)
            # we do not want to update the subscriber status
            if (
                esp_name == "Amazon SES"
                and event.esp_event.get("bounce", {}).get("bounceType") != "Permanent"
            ):
                campaign_status = CampaignSentStatusType.BLOCKED

            elif esp_name == "Postmark" and event.esp_event.get("Type") != "HardBounce":
                campaign_status = CampaignSentStatusType.BLOCKED

        if event.event_type == EventType.SENT:
            logger.debug(event.event_type + " : " + str(event.esp_event))
        else:
            logger.info(event.event_type + " : " + str(event.esp_event))

        with transaction.atomic():
            defaults = dict()
            if hasattr(subscriber_class().objects, "get_subscriber"):
                defaults["subscriber"] = subscriber_class().objects.get_subscriber(
                    event.recipient
                )
            c, is_create = CampaignSentEvent.objects.select_for_update().get_or_create(
                email=event.recipient,
                esp_message_id=event.message_id,
                defaults={**defaults},
            )
            if campaign_status is not None:
                c.result = campaign_status
            if event.event_type == EventType.OPENED:
                c.open_count = c.open_count + 1
            if event.event_type == EventType.CLICKED:
                c.click_count = c.click_count + 1
            c.save()

        update_subscriber(event.recipient, campaign_status)



from base64 import b64decode
from urllib.parse import urlparse, parse_qs

from PIL import Image
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.db.models import F
from django.http import HttpResponse, HttpResponseBadRequest, Http404
from django.http.request import validate_host
from django.shortcuts import redirect, get_object_or_404

from nuntius.models import MosaicoImage, CampaignSentEvent
from nuntius.utils import generate_placeholder, url_signature_is_valid


def mosaico_image_processor_view(request):
    if request.GET.get("method") == "placeholder":
        (width, height) = request.GET.get("params").split(",")
        image = generate_placeholder(int(width), int(height))
        response = HttpResponse(content_type="image/png")
        image.save(response, "PNG")
        return response

    if request.GET.get("src"):
        src = request.GET.get("src")
        host = urlparse(src).netloc.split(":")[0]
        allowed_hosts = settings.ALLOWED_HOSTS
        if settings.DEBUG and not allowed_hosts:
            allowed_hosts = ["localhost", "127.0.0.1", "[::1]"]
        if not validate_host(host, allowed_hosts):
            return HttpResponseBadRequest()

        (width, height) = request.GET.get("params").split(",")
        (width, height) = (
            int(width.replace("null", "0")),
            int(height.replace("null", "0")),
        )
        image = MosaicoImage.objects.get(
            file=urlparse(src).path.replace(settings.MEDIA_URL, "", 1)
        )
        image = Image.open(image.file.path)

        if width and not height:
            ratio = width / image.size[0]
        if height and not width:
            ratio = height / image.size[1]
        if width and height:
            ratio = min(width / image.size[0], height / image.size[1])

        image.resize((round(size * ratio) for size in image.size), Image.ANTIALIAS)
        response = HttpResponse(content_type=f"image/{image.format.lower()}")
        image.save(response, image.format)
        return response

    return HttpResponseBadRequest()


def track_open_view(request, tracking_id):
    CampaignSentEvent.objects.filter(tracking_id=tracking_id).update(
        open_count=F("open_count") + 1
    )
    return HttpResponse(
        b64decode(
            "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNkYAAAAAYAAjCB0C8AAAAASUVORK5CYII="
        ),
        content_type="image/png",
    )


def track_click_view(request, tracking_id, link, signature):
    campaign_sent_event = get_object_or_404(CampaignSentEvent, tracking_id=tracking_id)

    CampaignSentEvent.objects.filter(tracking_id=tracking_id).update(
        click_count=F("click_count") + 1
    )

    url = parse_qs("link=" + link).get("link")

    if url is None or len(url) != 1:
        raise

    if not url_signature_is_valid(campaign_sent_event.campaign, url[0], signature):
        raise PermissionDenied()

    return redirect(url[0])



import json

from django.conf import settings
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponseBadRequest, HttpResponse, JsonResponse
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.urls import reverse, path, resolve
from django.utils.html import format_html, format_html_join
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from django.views.decorators.csrf import ensure_csrf_cookie
from django.views.generic import CreateView

from nuntius.celery import nuntius_celery_app
from nuntius.models import Campaign, MosaicoImage, CampaignSentEvent
from nuntius._tasks import send_campaign
from nuntius.utils import NoCeleryError, build_absolute_uri


def subscriber_class():
    model = settings.NUNTIUS_SUBSCRIBER_MODEL
    return ContentType.objects.get(
        app_label=model.split(".")[0], model=model.split(".")[1].lower()
    ).model_class()


class MosaicoImageUploadView(CreateView):
    model = MosaicoImage
    fields = ("file",)

    def image_dict(self, image):
        return {
            "name": image.file.name,
            "size": image.file.size,
            "url": build_absolute_uri(self.request, image.file.url),
            "deleteUrl": build_absolute_uri(self.request, image.file.url),
            "deleteType": "DELETE",
            "thumbnailUrl": build_absolute_uri(self.request, image.file.thumbnail.url),
        }

    def form_invalid(self, form):
        return HttpResponseBadRequest()

    def form_valid(self, form):
        self.object = form.save()

        return JsonResponse({"files": [self.image_dict(self.object)]})

    def get(self, *args, **kwargs):
        return JsonResponse(
            {"files": [self.image_dict(image) for image in MosaicoImage.objects.all()]}
        )


class CampaignAdmin(admin.ModelAdmin):
    search_fields = ("name", "message_subject")
    autocomplete_fields = ("segment",)
    fieldsets = (
        (
            None,
            {
                "fields": (
                    "created",
                    "updated",
                    "name",
                    "message_from_name",
                    "message_from_email",
                    "message_reply_to_name",
                    "message_reply_to_email",
                    "message_subject",
                )
            },
        ),
        (
            _("Content"),
            {
                "fields": (
                    "available_variables",
                    "mosaico_buttons",
                    "message_content_text",
                )
            },
        ),
        (
            _("Sending details"),
            {
                "fields": (
                    "first_sent",
                    "segment",
                    "segment_subscribers",
                    "status",
                    "send_button",
                    "task_uuid",
                    "task_state",
                )
            },
        ),
        (
            _("Sending reports"),
            {
                "fields": (
                    "sent_to",
                    "sent_ok",
                    "sent_bounced",
                    "sent_complained",
                    "sent_blocked",
                    "unique_open_count",
                    "unique_click_count",
                    "open_count",
                    "click_count",
                )
            },
        ),
    )
    list_display = (
        "name",
        "message_subject",
        "segment",
        "status",
        "send_button",
        "sent_to",
    )
    list_filter = ("status",)
    readonly_fields = (
        "created",
        "updated",
        "first_sent",
        "segment_subscribers",
        "status",
        "send_button",
        "available_variables",
        "mosaico_buttons",
        "message_content_text",
        "sent_to",
        "sent_ok",
        "sent_bounced",
        "sent_complained",
        "sent_blocked",
        "unique_open_count",
        "unique_click_count",
        "open_count",
        "click_count",
        "task_uuid",
        "task_state",
    )
    save_as = True

    def get_object(self, request, object_id, from_field=None):
        object = super().get_object(request, object_id, from_field=from_field)
        try:
            self.task = object.get_task_and_update_status()
        except NoCeleryError:
            self.task = False

        return object

    def save_model(self, request, campaign, form, change):
        if "_saveasnew" in request.POST:
            original_pk = resolve(request.path).kwargs["object_id"]
            original_campaign = Campaign.objects.get(id=original_pk)
            campaign.message_content_html = original_campaign.message_content_html
            campaign.message_mosaico_data = original_campaign.message_mosaico_data

        return super().save_model(request, campaign, form, change)

    def segment_subscribers(self, instance):
        if instance.segment is None:
            return subscriber_class().objects.count()
        return instance.segment.get_subscribers_count()

    segment_subscribers.short_description = _("Subscribers")

    def sent_to(self, instance):
        return format_html(
            '<a href="{}">{}</a>',
            reverse("admin:nuntius_campaignsentevent_changelist")
            + "?campaign_id__exact="
            + str(instance.pk),
            str(instance.get_sent_count()),
        )

    sent_to.short_description = _("Sent to")

    def sent_ok(self, instance):
        return instance.get_ok_count()

    sent_ok.short_description = _("Ok")

    def sent_bounced(self, instance):
        return instance.get_bounced_count()

    sent_bounced.short_description = _("Bounced")

    def sent_complained(self, instance):
        return instance.get_complained_count()

    sent_complained.short_description = _("Complained")

    def sent_blocked(self, instance):
        return instance.get_blocked_count()

    sent_blocked.short_description = _("Blocked")

    def open_count(self, instance):
        return instance.get_open_count()

    open_count.short_description = _("Open count")

    def unique_open_count(self, instance):
        return instance.get_unique_open_count()

    unique_open_count.short_description = _("Unique open count")

    def click_count(self, instance):
        return instance.get_click_count()

    click_count.short_description = _("Click count")

    def unique_click_count(self, instance):
        return instance.get_unique_click_count()

    unique_click_count.short_description = _("Unique click count")

    def task_state(self, instance):
        task = self.task
        if task:
            return task[0]
        if instance.task_uuid is None:
            return "-"

        if task is False:
            return _("Connection to celery failed")

        return _(
            "A sending task has been scheduled, is not launched yet, or has been stopped for unkown reasons."
        )

    task_state.short_description = _("Send task state")

    def send_button(self, instance):
        if instance.pk is None:
            return mark_safe("-")
        if instance.status == Campaign.STATUS_SENDING:
            return format_html(
                '<a href="{}" class="button">' + _("Pause") + "</a>",
                reverse("admin:nuntius_campaign_pause", args=[instance.pk]),
            )
        if instance.task_uuid is None:
            return format_html(
                '<a href="{}" class="button">' + _("Send") + "</a>",
                reverse("admin:nuntius_campaign_send", args=[instance.pk]),
            )

        return mark_safe("-")

    send_button.short_description = _("Send")

    def available_variables(self, instance):
        if instance.segment is not None:
            qs = instance.segment.get_subscribers_queryset()
        else:
            qs = subscriber_class().objects.all()

        data = qs.first().get_subscriber_data()

        return format_html_join(
            mark_safe("<br/>"), "<b>{{{{ {} }}}}</b> ({})", data.items()
        )

    available_variables.short_description = _(
        "Available variables (for first subscriber)"
    )

    def mosaico_buttons(self, instance):
        if instance.pk is None:
            return mark_safe("-")

        if not instance.message_mosaico_data:
            default_template = (
                settings.STATIC_URL
                + "/nuntius/templates/versafix-1/template-versafix-1.html",
                _("Default template"),
            )
            templates = getattr(
                settings, "NUNTIUS_MOSAICO_TEMPLATES", [default_template]
            )

            return format_html_join(
                " ",
                '<a href="{}" class="button">{}</a>',
                (
                    (
                        reverse("admin:nuntius_campaign_mosaico", args=[instance.pk])
                        + "#"
                        + template,
                        name,
                    )
                    for template, name in templates
                ),
            )

        return format_html(
            '<a href="{}" class="button">' + _("Access the editor") + "</a> "
            '<a href="{}" class="button" target="_blank">'
            + _("Preview result")
            + "</a>",
            reverse("admin:nuntius_campaign_mosaico", args=[instance.pk]),
            reverse("admin:nuntius_campaign_mosaico_preview", args=[instance.pk]),
        )

    mosaico_buttons.short_description = _("HTML content")

    def get_urls(self):
        return [
            path(
                "<pk>/send/",
                self.admin_site.admin_view(self.send_view),
                name="nuntius_campaign_send",
            ),
            path(
                "<pk>/pause/",
                self.admin_site.admin_view(self.pause_view),
                name="nuntius_campaign_pause",
            ),
            path(
                "<pk>/mosaico/",
                ensure_csrf_cookie(self.admin_site.admin_view(self.mosaico_view)),
                name="nuntius_campaign_mosaico",
            ),
            path(
                "<pk>/mosaico/preview/",
                self.admin_site.admin_view(self.mosaico_preview),
                name="nuntius_campaign_mosaico_preview",
            ),
            path(
                "<pk>/mosaico/save/",
                self.admin_site.admin_view(self.mosaico_save_view),
                name="nuntius_campaign_mosaico_save",
            ),
            path(
                "<pk>/mosaico/data/",
                self.admin_site.admin_view(self.mosaico_load_view),
                name="nuntius_campaign_mosaico_load",
            ),
            path(
                "<pk>/mosaico/upload/",
                self.admin_site.admin_view(MosaicoImageUploadView.as_view()),
                name="nuntius_campaign_mosaico_image_upload",
            ),
        ] + super().get_urls()

    def send_view(self, request, pk):
        campaign = Campaign.objects.get(pk=pk)
        if campaign.task_uuid is not None:
            return redirect(reverse("admin:nuntius_campaign_change", args=[pk]))

        r = send_campaign.delay(pk, build_absolute_uri(request, location="/")[:-1])
        campaign.task_uuid = r.id
        campaign.save(update_fields=["task_uuid"])

        return redirect(reverse("admin:nuntius_campaign_change", args=[pk]))

    def pause_view(self, request, pk):
        campaign = Campaign.objects.get(pk=pk)
        if campaign.task_uuid is None:
            return redirect(reverse("admin:nuntius_campaign_change", args=[pk]))

        nuntius_celery_app.control.revoke(str(campaign.task_uuid), terminate=True)

        campaign.task_uuid = None
        campaign.save(update_fields=["task_uuid"])

        return redirect(reverse("admin:nuntius_campaign_change", args=[pk]))

    def mosaico_view(self, request, pk):
        return TemplateResponse(
            request=request,
            template="nuntius/editor.html",
            context={
                "image_processor_backend_url": build_absolute_uri(
                    request, reverse("nuntius_mosaico_image_processor")
                ),
                "save_url": reverse("admin:nuntius_campaign_mosaico_save", args=[pk]),
                "image_upload_url": reverse(
                    "admin:nuntius_campaign_mosaico_image_upload", args=[pk]
                ),
                "load_data_url": reverse(
                    "admin:nuntius_campaign_mosaico_load", args=[pk]
                ),
                "change_url": reverse("admin:nuntius_campaign_change", args=[pk]),
            },
        )

    def mosaico_preview(self, request, pk):
        campaign = Campaign.objects.get(pk=pk)

        return HttpResponse(campaign.message_content_html)

    def mosaico_load_view(self, request, pk):
        campaign = Campaign.objects.get(pk=pk)

        return JsonResponse(json.loads(campaign.message_mosaico_data or "{}"))

    def mosaico_save_view(self, request, pk):
        if request.method != "POST":
            return HttpResponseBadRequest()

        html = request.POST.get("html")
        metadata = request.POST.get("metadata")
        content = request.POST.get("content")

        campaign = Campaign.objects.get(pk=pk)
        campaign.message_content_html = html
        campaign.message_mosaico_data = json.dumps(
            {"metadata": json.loads(metadata), "content": json.loads(content)}
        )
        campaign.save()

        return HttpResponse()


class TrackingFilter(admin.SimpleListFilter):
    title = _("Tracking")
    parameter_name = "tracking"

    def lookups(self, request, model_admin):
        return ("O", _("Opened")), ("C", _("Clicked"))

    def queryset(self, request, queryset):
        if self.value() == "O":
            return queryset.exclude(open_count=0)
        if self.value() == "C":
            return queryset.exclude(click_count=0)


class CampaignSentEventAdmin(admin.ModelAdmin):
    def has_change_permission(self, *args, **kwargs):
        return False

    def has_add_permission(self, *args, **kwargs):
        return False

    actions = None
    readonly_fields = ("subscriber_filter", "campaign_filter")
    list_filter = ("result", TrackingFilter)
    list_display_links = None

    def get_list_display(self, request):
        list_display = ("email", "datetime", "result", "open_count", "click_count")
        if request.GET.get("campaign_id__exact") is None:
            list_display = ("campaign_filter", *list_display)
        if request.GET.get("subscriber_id__exact") is None:
            list_display = ("subscriber_filter", *list_display)

        return list_display

    def subscriber_filter(self, instance):
        if instance.subscriber is None:
            return "-"

        return format_html(
            '<a href="{}">{}</a>',
            reverse("admin:nuntius_campaignsentevent_changelist")
            + "?subscriber_id__exact="
            + str(instance.subscriber_id),
            str(instance.subscriber),
        )

    subscriber_filter.short_description = _("Subscriber")

    def campaign_filter(self, instance):
        if instance.campaign is None:
            return "-"

        return format_html(
            "<a href={}>{}</a>",
            reverse("admin:nuntius_campaignsentevent_changelist")
            + "?campaign_id__exact="
            + str(instance.campaign_id),
            str(instance.campaign),
        )

    campaign_filter.short_description = _("Campaign")

    def changelist_view(self, request, extra_context=None):
        title = _("Sent events")
        campaign, subscriber = (None, None)

        if request.GET.get("campaign_id__exact") is not None:
            campaign = Campaign.objects.filter(
                id=request.GET.get("campaign_id__exact")
            ).first()
        if request.GET.get("subscriber_id__exact") is not None:
            subscriber = (
                subscriber_class()
                .objects.filter(id=request.GET.get("subscriber_id__exact"))
                .first()
            )

        if campaign and subscriber:
            title = _(
                f"Sent event for campaign %(campaign)s and subscriber %(subscriber)s"
            ) % {"campaign": str(campaign), "subscriber": str(subscriber)}
        elif campaign:
            title = mark_safe(
                _("Sent events for campaign %s")
                % (
                    format_html(
                        '<a href="{}">{}</a>',
                        reverse("admin:nuntius_campaign_change", args=[campaign.pk]),
                        str(campaign),
                    ),
                )
            )
        elif subscriber:
            title = _("Sent events for subscriber %s") % (str(subscriber),)

        return super().changelist_view(request, extra_context={"title": title})


if not getattr(settings, "NUNTIUS_DISABLE_DEFAULT_ADMIN", False):
    admin.site.register(Campaign, CampaignAdmin)
    admin.site.register(CampaignSentEvent, CampaignSentEventAdmin)



"""
    Nuntius needs its own celery app. This module is named _tasks so it is not autodiscovered by other celery app that
    live in the project.
"""

import re
from smtplib import SMTPServerDisconnected, SMTPRecipientsRefused
from time import sleep
from urllib.parse import quote as url_quote

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core import mail
from django.core.mail import EmailMultiAlternatives
from django.db import transaction
from django.template import Template, Context
from django.urls import reverse
from django.utils import timezone

from nuntius.celery import nuntius_celery_app
from nuntius.models import (
    Campaign,
    CampaignSentEvent,
    CampaignSentStatusType,
    AbstractSubscriber,
)
from nuntius.utils import sign_url

try:
    from anymail.exceptions import AnymailRecipientsRefused
except:

    class AnymailRecipientsRefused(BaseException):
        pass


def replace_url(url, campaign, tracking_id, public_url):
    return public_url + reverse(
        "nuntius_track_click",
        kwargs={
            "tracking_id": tracking_id,
            "signature": sign_url(campaign, url),
            "link": url_quote(url, safe=""),
        },
    )


def replace_vars(campaign, data, public_url):
    var_regex = re.compile(r"\[([-a-zA-Z_]+)\]")
    context = Context(data)

    html_rendered_content = Template(
        var_regex.sub(r"{{ \1 }}", campaign.message_content_html)
    ).render(context=context)

    html_rendered_content = re.sub(
        r"(<a[^>]* href\s*=[\s\"']*)(http[^\"'>\s]+)",
        lambda match: match.group(1)
        + replace_url(
            match.group(2), campaign, data["nuntius_tracking_id"], public_url
        ),
        html_rendered_content,
        flags=re.MULTILINE | re.IGNORECASE,
    )

    text_rendered_content = Template(
        var_regex.sub(r"{{ \1 }}", campaign.message_content_text)
    ).render(context=context)

    return text_rendered_content, html_rendered_content


def reset_connection(connection):
    try:
        connection.close()
    except Exception:
        pass

    for retry in reversed(range(1, 11)):
        try:
            sleep(1 / retry)
            connection.open()
        except Exception:
            connection.close()
        else:
            break


def insert_tracking_image(public_url, html_message):
    img_url = (
        public_url + reverse("nuntius_mount_path") + "open/{{ nuntius_tracking_id }}"
    )
    img = '<img src="{}" width="1" height="1" alt="nt">'.format(img_url)
    return re.sub(
        r"(<\/body\b)", img + r"\1", html_message, flags=re.MULTILINE | re.IGNORECASE
    )


@nuntius_celery_app.task()
def send_campaign(campaign_pk, public_url):
    campaign = Campaign.objects.get(pk=campaign_pk)
    campaign.status = Campaign.STATUS_SENDING
    if campaign.first_sent is None:
        campaign.first_sent = timezone.now()
    campaign.save(update_fields=["status", "first_sent"])

    if campaign.segment is None:
        model = settings.NUNTIUS_SUBSCRIBER_MODEL
        model_class = ContentType.objects.get(
            app_label=model.split(".")[0], model=model.split(".")[1].lower()
        ).model_class()
        queryset = model_class.objects.all()
    else:
        queryset = campaign.segment.get_subscribers_queryset()

    campaign.message_content_html = insert_tracking_image(
        public_url, campaign.message_content_html
    )

    def send_message(connection, sent_event, message, retries=10):
        try:
            message.send()
        except SMTPServerDisconnected as e:
            if retries == 0:
                campaign.status = Campaign.STATUS_ERROR
                campaign.save(update_fields=["status"])
                raise e

            reset_connection(connection)

            sleep(1 / retries)

            send_message(connection, sent_event, message, retries=retries - 1)
        except (SMTPRecipientsRefused, AnymailRecipientsRefused):
            sent_event.result = CampaignSentStatusType.BLOCKED
            sent_event.save()
        else:
            if hasattr(message, "anymail_status"):
                if message.anymail_status.recipients[sent_event.email].status in [
                    "invalid",
                    "rejected",
                    "failed",
                ]:
                    sent_event.result = CampaignSentStatusType.REJECTED
                else:
                    sent_event.result = CampaignSentStatusType.UNKNOWN
                sent_event.esp_message_id = message.anymail_status.recipients[
                    sent_event.email
                ].message_id
            else:
                sent_event.result = CampaignSentStatusType.UNKNOWN
            sent_event.save()

    try:
        with mail.get_connection(
            backend=getattr(settings, "NUNTIUS_EMAIL_BACKEND", None)
        ) as connection:
            for subscriber in queryset.iterator():
                if (
                    subscriber.get_subscriber_status()
                    != AbstractSubscriber.STATUS_SUBSCRIBED
                ):
                    continue

                email = subscriber.get_subscriber_email()

                (sent_event, created) = CampaignSentEvent.objects.get_or_create(
                    campaign=campaign, subscriber=subscriber, email=email
                )

                if sent_event.result != CampaignSentStatusType.PENDING:
                    continue

                with transaction.atomic():
                    sent_event = CampaignSentEvent.objects.select_for_update().get(
                        subscriber=subscriber, campaign=campaign
                    )

                    if sent_event.result != CampaignSentStatusType.PENDING:
                        continue

                    subscriber_data = {
                        "nuntius_tracking_id": sent_event.tracking_id,
                        **subscriber.get_subscriber_data(),
                    }

                    from_email = (
                        f"{campaign.message_from_name} <{campaign.message_from_email}>"
                        if campaign.message_from_name
                        else campaign.message_from_email
                    )
                    text_body, html_body = replace_vars(
                        campaign, subscriber_data, public_url
                    )
                    message = EmailMultiAlternatives(
                        subject=campaign.message_subject,
                        body=text_body,
                        from_email=from_email,
                        to=[email],
                        reply_to=campaign.message_reply_to_email,
                        connection=connection,
                    )
                    message.attach_alternative(html_body, "text/html")
                    send_message(connection, sent_event, message)

            campaign.status = Campaign.STATUS_SENT
            campaign.task_uuid = None
            campaign.save(update_fields=["status", "task_uuid"])
    except Exception as e:
        campaign.status = Campaign.STATUS_ERROR
        campaign.task_uuid = None
        campaign.save(update_fields=["status", "task_uuid"])
        raise e



from django.urls import path, reverse_lazy
from django.views.generic import RedirectView

from nuntius.views import (
    mosaico_image_processor_view,
    track_open_view,
    track_click_view,
)

urlpatterns = [
    path(
        "",
        RedirectView.as_view(url=reverse_lazy("admin:nuntius")),
        name="nuntius_mount_path",
    ),
    path("img/", mosaico_image_processor_view, name="nuntius_mosaico_image_processor"),
    path("open/<str:tracking_id>", track_open_view, name="nuntius_track_open"),
    path(
        "link/<str:tracking_id>/<str:link>/<str:signature>",
        track_click_view,
        name="nuntius_track_click",
    ),
]



# Generated by Django 2.2.5 on 2019-10-22 07:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("nuntius", "0015_only_one_segment_model")]

    operations = [
        migrations.AlterModelOptions(
            name="campaignsentevent",
            options={
                "ordering": ["-datetime"],
                "verbose_name": "Sent event",
                "verbose_name_plural": "Sent events",
            },
        ),
        migrations.AlterField(
            model_name="campaignsentevent",
            name="campaign",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="nuntius.Campaign",
                verbose_name="Campaign",
            ),
        ),
        migrations.AlterField(
            model_name="campaignsentevent",
            name="datetime",
            field=models.DateTimeField(
                auto_now_add=True, db_index=True, verbose_name="Sending time"
            ),
        ),
        migrations.AlterField(
            model_name="campaignsentevent",
            name="email",
            field=models.EmailField(
                db_index=True,
                max_length=254,
                verbose_name="Email address at sending time",
            ),
        ),
        migrations.AddIndex(
            model_name="campaignsentevent",
            index=models.Index(
                fields=["email", "datetime"], name="nuntius_cam_email_04da96_idx"
            ),
        ),
    ]



from datetime import timedelta

from django.conf import settings
from django.utils.timezone import now

from nuntius.admin import subscriber_class
from nuntius.models import CampaignSentStatusType, AbstractSubscriber, CampaignSentEvent

successful_sent = (CampaignSentStatusType.UNKNOWN, CampaignSentStatusType.OK)


def update_subscriber(email, campaign_status):
    statuses = {
        CampaignSentStatusType.BOUNCED: None,
        CampaignSentStatusType.UNSUBSCRIBED: AbstractSubscriber.STATUS_UNSUBSCRIBED,
        CampaignSentStatusType.COMPLAINED: AbstractSubscriber.STATUS_COMPLAINED,
    }

    if campaign_status not in statuses:
        return

    model_class = subscriber_class()

    # if unsubscribe or spam, we change the status of the subscriber
    if campaign_status != CampaignSentStatusType.BOUNCED:
        model_class.objects.set_subscriber_status(email, statuses.get(campaign_status))
        return

    email_events = CampaignSentEvent.objects.filter(email=email)

    # if first email sent is a bounce, we bounce forever
    if not email_events.filter(result__in=successful_sent).exists():
        model_class.objects.set_subscriber_status(
            email, AbstractSubscriber.STATUS_BOUNCED
        )
        return

    default_params = {"consecutive": 1, "duration": 7, "limit": 3}
    params = {**default_params, **getattr(settings, "NUNTIUS_BOUNCE_PARAMS", dict())}

    max_bounce_duration_ago = now() - timedelta(days=params["duration"])

    recent_successful_sent = email_events.filter(
        result__in=successful_sent, datetime__gt=max_bounce_duration_ago
    ).exists()

    # if there is at least a successful sending in `duration`, it is allowed up to `limit`
    if (
        recent_successful_sent
        and email_events.filter(
            result=CampaignSentStatusType.BOUNCED, datetime__gt=max_bounce_duration_ago
        ).count()
        <= params["limit"]
    ):
        return

    # it is also ok if we have at least a successful sending in last `consecutive` + 1
    for sent_event in email_events[: params["consecutive"] + 1]:
        if sent_event.result in successful_sent:
            return

    # in all other case, we bounce
    model_class.objects.set_subscriber_status(email, AbstractSubscriber.STATUS_BOUNCED)



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "README.md")) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="nuntius",
    version=__import__("nuntius").__version__,
    packages=find_packages(exclude=["tests", "tests.*"]),
    include_package_data=True,
    license="GPLv3",
    description="A model agnostic Django newsletter app integrating Mosaico.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/lafranceinsoumise/nuntius",
    author="Jill Royer",
    python_requires="~=3.6",
    author_email="perso@jillroyer.me",
    install_requires=("celery", "pillow", "html2text"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Communications :: Email",
    ],
)



# Generated by Django 2.1.5 on 2019-02-05 10:47

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("nuntius", "0005_change_status_choices")]

    operations = [
        migrations.AlterField(
            model_name="campaignsentevent",
            name="subscriber",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.NUNTIUS_SUBSCRIBER_MODEL,
                verbose_name="Subscriber",
            ),
        )
    ]
